/*
    RGB Equalizer Filter for VirtualDub (2.0).
	Copyright (C) 2004-2005 Eugene Khoroshavin

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	Eugene Khoroshavin
	fdump@mail.ru
*/

#include <windows.h>
#include <commctrl.h>
#include <stdio.h>
#include <crtdbg.h>
#include <math.h>
#include <stdlib.h>

#include "ScriptInterpreter.h"
#include "ScriptError.h"
#include "ScriptValue.h"

#include "resource.h"
#include "filter.h"

///////////////////////////////////////////////////////////////////////////

int RunProc(const FilterActivation *fa, const FilterFunctions *ff);
int StartProc(FilterActivation *fa, const FilterFunctions *ff);
int EndProc(FilterActivation *fa, const FilterFunctions *ff);
long ParamProc(FilterActivation *fa, const FilterFunctions *ff);
int InitProc(FilterActivation *fa, const FilterFunctions *ff);
int ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd);
void StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str);
void ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc);
bool FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen);
UCHAR mU(long r, long g, long b, long n);
void keepU(long *r, long *g, long *b, UCHAR u, long n);
void normrgb(long *r, long *g, long *b);
///////////////////////////////////////////////////////////////////////////
#define max3(a,b,c)            max(max(a,b),c)
#define min3(a,b,c)            min(min(a,b),c)
 
typedef struct MyFilterData {
	IFilterPreview		*ifp;

	long			lv[10][20];	//sliders vals
	long			Y;			//luma-abstraction
	long			Flg[10];	/*
								0x00000001
								..........
								0x00010000		EQ sliders
								0x00020000		keep luma
								0x00040000		Lock sliders
								*/
	long			nmfunc;

	long			todoFlg;		/*
					0x00000001		ON - OFF
					0x00000002		hue do
					0x00000004		rgb do
					0x00000008		saturation do
					0x00000010		Sharp<->Smooth do	*/
	UCHAR			rtbl[256];
	UCHAR			gtbl[256];
	UCHAR			btbl[256];
	long			incs[256];
	long			incr[256];
	long			incg[256];
	long			incb[256];
	long			incrk[256];
	long			incgk[256];
	long			incbk[256];
	long			hue[256];
	long			shsm[256];
} MyFilterData;

////////////////////////////////////////////////////////////////////////////////
void setFlg (MyFilterData *mfd, long ns);
///////////////////////////////////////////////////////////////////////////////

bool FssProc(FilterActivation *fa, const FilterFunctions *ff, char *buf, int buflen) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	int i, k, l;
	char mary[500];
	long tval;

	for (i=1; i<500; i++) 
		mary[i]=(char)0x80; // not use 8th bit
	mary[0] = 34;
	i = 1;
	for (k=0; k<10; k++)
		for (l=0; l<20; l++) {
			tval = mfd->lv[k][l] + 128;
			mary[i++] |= (char)(tval & 0x7f);
			mary[i++] |= (char)((tval>>7) & 0x7f);
		}
	tval = mfd->Y + 128;
	mary[i++] |= (char)(tval & 0x7f);
	mary[i++] |= (char)((tval>>7) & 0x7f);
	mary[i++] |= (char)(mfd->nmfunc);
	for (k=0; k<10; k++) 
		for (l=0; l<4; l++)
			mary[i++] |= (char)(mfd->Flg[k] >> (l*7));
	mary[i++] = 34;
	mary[i++] = 0;

	_snprintf(buf, buflen, "Config(%s)", mary, i);

	return true;
}

void ScriptConfig(IScriptInterpreter *isi, void *lpVoid, CScriptValue *argv, int argc) {
	FilterActivation *fa = (FilterActivation *)lpVoid;
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	int i, k, l;
	char **mar = argv[0].asString();
	char *mary = (char *)*mar;

	i=0;
	for (k=0; k<10; k++)
		for (l=0; l<20; l++) {
			mfd->lv[k][l] = mary[i++] & 0x7f;
			mfd->lv[k][l] |= (((long)(mary[i++] & 0x7f))<<7);
			mfd->lv[k][l] -= 128;
		}
	mfd->Y = mary[i++] & 0x7f;
	mfd->Y |= (((long)(mary[i++] & 0x7f))<<7);
	mfd->Y -= 128;
	mfd->nmfunc = mary[i++] & 0x7f;
	for (k=0; k<10; k++) {
		mfd->Flg[k] = 0;
		for (l=0; l<4; l++) 
			mfd->Flg[k] |= ((long)(mary[i++] & 0x7f)) << l*7;
	}
	setFlg(mfd, 255);
}

ScriptFunctionDef func_defs[]={
	{ (ScriptFunctionPtr)ScriptConfig, "Config", "0s" },
	{ NULL },
};

CScriptObject script_obj={
	NULL, func_defs
};

struct FilterDefinition filterDef_tutorial = {

	NULL, NULL, NULL,		// next, prev, module
	"RGB Equalizer(2.0)",	// name
	"17-channels equalizer(RGB levels & gamma, hue, saturation, sharpness<->smoothness).",
							// desc
	"Eugene Khoroshavin", 		// maker
	NULL,					// private_data
	sizeof(MyFilterData),	// inst_data_size

	InitProc,				// initProc
	NULL,					// deinitProc
	RunProc,				// runProc
	NULL,					// paramProc
	ConfigProc, 			// configProc
	StringProc, 			// stringProc
	StartProc,				// startProc
	EndProc,				// endProc

	&script_obj,			// script_obj
	FssProc,				// fssProc

};

int StartProc(FilterActivation *fa, const FilterFunctions *ff)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	return 0;
}

int EndProc(FilterActivation *fa, const FilterFunctions *ff)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	return 0;
}

///////////////////////////////////////////////////////////////////////////

int RunProc(const FilterActivation *fa, const FilterFunctions *ff)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	const long	width = fa->src.w;
	const long	height = fa->src.h;
	Pixel32 *src, *dst;
	long r, g, b, w, h;
	UCHAR UU;
	long salign = fa->src.modulo;
	long dalign = fa->dst.modulo;
	const long	vol = (fa->src.w + 1) * fa->src.h;
	UCHAR *myY = new UCHAR[vol];
	long huedo =	mfd->todoFlg & 0x02;
	long rgbdo =	mfd->todoFlg & 0x04;
	long satdo =	mfd->todoFlg & 0x08;
	long shsmdo =	(mfd->todoFlg & 0x10) > 0;
	long keeph =	mfd->Flg[0] & 0x00020000;
	long keeprgb =	mfd->Flg[1] & 0x00020000 |
					mfd->Flg[2] & 0x00020000 |
					mfd->Flg[3] & 0x00020000 |
					mfd->Flg[4] & 0x00020000 |
					mfd->Flg[5] & 0x00020000 |
					mfd->Flg[6] & 0x00020000;
	long fRd[] = {128,128,128,128};
	long nRd;
	if (mfd->lv[9][0] > 0) {
		nRd = 2 + (mfd->lv[9][0] * 2 - 1)/128;
		fRd[nRd] = mfd->lv[9][0] * 2 - (nRd - 2)*128;
	}
	else {
		nRd = 1;
		fRd[1] = 128 + mfd->lv[9][0];
	}

	long pos=0;
	src = fa->src.data;
	dst = fa->dst.data;
	h = height;
	if(mfd->todoFlg & 0x01) {
		pos=0;
		src = fa->src.data;
		dst = fa->dst.data;
		h = height;
		do {
			w = width;
			do {
				b = mfd->btbl[((UCHAR *)src)[0]];
				g = mfd->gtbl[((UCHAR *)src)[1]];
				r = mfd->rtbl[((UCHAR *)src)[2]];
				UU = mU(r, g, b, mfd->Y);
				if (huedo) {	//Hue
					long S, H, B;
					B = min3(r,g,b);
					S = max3(r,g,b) - B;
					if(S) {
						if(B == g) {
							if(r >= b) H = ((b - B)<<6)/S;
							else H = ((2 * S - r + B)<<6)/S;
						}
						else {
							if(B == r) {
								if(b >= g) H = ((2 * S + g - B)<<6)/S;
								else H = ((4 * S - b + B)<<6)/S;
							}
							else {
								if(g >= r) H = ((4 * S + r - B)<<6)/S;
								else H = ((6 * S - g + B)<<6)/S;
							}
						}
						H += mfd->hue[UU];
						if (H>384) H -= 384;
						if (H<0) H += 384;
						if (H==384) H=0;
						long N = H>>6;
						switch (N) {
						case 0:
							r = B + S;
							g = B;
							b = B + ((H * S)>>6);
							break;
						case 1:
							r = B + (((128- H) * S)>>6);
							g = B;
							b = B + S;
							break;
						case 2:
							r = B;
							g = B + (((H - 128) * S)>>6);
							b = B + S;
							break;
						case 3:
							r = B;
							g = B + S;
							b = B + (((256- H) * S)>>6);
							break;
						case 4:
							r = B + (((H - 256) * S)>>6);
							g = B + S;
							b = B;
							break;
						case 5:
							r = B + S;
							g = B + (((384 - H) * S)>>6);
							b = B;
							break;
						default:
							r = 0;
							g = 255;
							b = 255;
							break;
						}
						if(keeph) keepU(&r, &g, &b, UU, mfd->Y);
					}
				}
				if (rgbdo) {	//RGB
					if(mfd->Y==6){
						r += mfd->incrk[r];
						g += mfd->incgk[g];
						b += mfd->incbk[b];
						if (keeprgb) keepU(&r, &g, &b, UU, mfd->Y);
						r += mfd->incr[r];
						g += mfd->incg[g];
						b += mfd->incb[b];
						normrgb(&r, &g, &b);
					}
					else {
						r += mfd->incrk[UU];
						g += mfd->incgk[UU];
						b += mfd->incbk[UU];
						if (keeprgb) keepU(&r, &g, &b, UU, mfd->Y);
						r += mfd->incr[UU];
						g += mfd->incg[UU];
						b += mfd->incb[UU];
						normrgb(&r, &g, &b);
					}
				}
				if (satdo) {	//Saturation
					long U = mU(r, g, b, mfd->Y);
					r = U + (((r - U) * mfd->incs[U])>>7);
					g = U + (((g - U) * mfd->incs[U])>>7);
					b = U + (((b - U) * mfd->incs[U])>>7);
					normrgb(&r, &g, &b);
				}
				if(shsmdo) myY[pos] = mU(r, g, b, mfd->Y);
				dst[0] = ((int) r << 16) | ((int) g << 8) | ((int) b);
				src++;
				dst++;
				pos++;
			} while(--w);
			src	= (Pixel32 *)((char *)src + salign);
			dst	= (Pixel32 *)((char *)dst + dalign);
		} while(--h);
/////////////////////////////////////////
		if(shsmdo) {	//Sharp<->Smooth
			pos=0;
			dst = fa->dst.data;
			h = height;
			do {
				w = width;
				do {
					UU = myY[pos];
					if (mfd->shsm[UU] != 0) {
						long nn = 0;
						b = ((UCHAR *)dst)[0];
						g = ((UCHAR *)dst)[1];
						r = ((UCHAR *)dst)[2];
						long dU = 0;
						long yy, xx, adr;
						if(mfd->lv[9][0] > 0) {
							for (yy = -nRd; yy <= nRd; yy ++) {
								for (xx = -nRd; xx <= nRd; xx++) {
									adr = pos + yy * width + xx;
									if ((adr >= 0) && (adr < vol) && (adr != pos)) {
										long rr = fRd[max(abs(yy),abs(xx))];
										nn += rr;
										dU += ((UU - myY[adr]) * rr);
									}
								}
							}
						}
						else {
							for (yy = -1; yy <= 1; yy ++) {
								for (xx = -1; xx <= 1; xx++) {
									adr = pos + yy * width + xx;
									if ((adr >= 0) && (adr < vol) && (adr != pos)) {
										nn += 128;
										dU += ((UU - myY[adr]) * fRd[max(abs(yy),abs(xx))]);
									}
								}
							}
						}
						if (nn) {
							if (mfd->shsm[UU] > 0) {
								nn <<= 5;
								dU = (dU * mfd->shsm[UU])/nn;
							}
							else {
								nn <<= 7;
								dU = (dU * mfd->shsm[UU])/nn;
							}
							r += dU;
							g += dU;
							b += dU;
							normrgb(&r, &g, &b);
						}
					dst[0] = ((int) r << 16) | ((int) g << 8) | ((int) b);
					}
					dst++;
					pos++;
				} while(--w);
				dst	= (Pixel32 *)((char *)dst + dalign);
			} while(--h);
		}
//////////////////////////////////////////
	}
	else {
		do {
			w = width;
			do {
				dst[0] = src[0];
				src++;
				dst++;
			} while(--w);
			src	= (Pixel32 *)((char *)src + salign);
			dst	= (Pixel32 *)((char *)dst + dalign);
		} while(--h);
	}
	delete myY;
	return 0;
}

extern "C" int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat);
extern "C" void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff);

static FilterDefinition *fd_tutorial;

int __declspec(dllexport) __cdecl VirtualdubFilterModuleInit2(FilterModule *fm, const FilterFunctions *ff, int& vdfd_ver, int& vdfd_compat) {
	if (!(fd_tutorial = ff->addFilter(fm, &filterDef_tutorial, sizeof(FilterDefinition))))
		return 1;

	vdfd_ver = VIRTUALDUB_FILTERDEF_VERSION;
	vdfd_compat = VIRTUALDUB_FILTERDEF_COMPATIBLE;
	return 0;
}

void __declspec(dllexport) __cdecl VirtualdubFilterModuleDeinit(FilterModule *fm, const FilterFunctions *ff) {
	ff->removeFilter(fd_tutorial);
}

int InitProc(FilterActivation *fa, const FilterFunctions *ff) {
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	int k,l;
	for (k=0; k<10; k++)
		for (l=0; l<20; l++)
			mfd->lv[k][l] = 0;
	mfd->Y = 0;
	mfd->nmfunc = 1;
	mfd->Flg[0] = 0;
	for (k=1;k<9;k++)
		mfd->Flg[k] = 0x010101;
	mfd->Flg[9] = 0x010000;
	mfd->todoFlg = 0x01;
	setFlg(mfd, 255);
	return 0;
}

BOOL CALLBACK ConfigDlgProc(HWND hdlg, UINT msg, WPARAM wParam, LPARAM lParam)
{
	MyFilterData *mfd = (MyFilterData *)GetWindowLong(hdlg, DWL_USER);
	HWND vhWnd[20];
	int i, nf;
	char * nult = "";
	int winst[] = {SW_HIDE, SW_SHOW};
	static const int fufl[] = {0x01, 0x02, 0x04, 0x08, 0x0c, 0x0a, 0x06, 0x0e, 0x10, 0x20};
	static const char *txt1[] = {
		"Hue",
		"Red",
		"Green",
		"Blue",
		"Cyan",
		"Magenta",
		"Yellow",
		"RGB",
		"Saturation",
		"Sharp<->Smooth"	};
	static const char *txt2[] = {
		"Hue",
		"Gamma R",
		"Gamma G",
		"Gamma B",
		"Gamma GB (Cyan)",
		"Gamma RB (Magenta)",
		"Gamma RG (Yellow)",
		"Gamma RGB",
		"Saturation",
		"Radius"	};
	static const char *txt3[] = {"OFF","ON"};
	static const int keponf[] = {1,1,1,1,1,1,1,0,0,0};
	static const char *yvars[] = {
		"(5r+8g+3b)/16",
		"(r+g+b)/3",
		"0.299r+0.587g+0.114b",
		"0.2126r+0.7152g+0.0722b",
		"Max(r,g,b) - V",
		"(Max+Min)/2 - L",
		"Separately r, g, b"};
	static const char *eqs[] = {
		"Single",
		"EQ-3",
		"EQ-5",
		"EQ-9",
		"EQ-17",
		"Dark",
		"Light"};
	static const long eqsf[] = {0x010000, 0x010101, 0x011111, 0x015555, 0x01FFFF, 0x010011, 0x011001};

	for (i=0;i<20;i++)
		vhWnd[i] = GetDlgItem(hdlg, IDC_GAMMA + i);

	switch(msg) {
		case WM_INITDIALOG:
			SetWindowLong(hdlg, DWL_USER, lParam);
			mfd = (MyFilterData *)lParam;

			setFlg(mfd, 255);
			for (i = 0; i < 7; i++) 
				SendMessage(GetDlgItem(hdlg, IDC_LUMA), CB_ADDSTRING, 0, (LPARAM)(LPCTSTR)yvars[i]);
			for (i = 0; i < 7; i++) 
				SendMessage(GetDlgItem(hdlg, IDC_PRESET), CB_ADDSTRING, 0, (LPARAM)(LPCTSTR)eqs[i]);
			for (i = 0; i < 19; i++) {
				SendMessage(vhWnd[i], TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-128, 128));
				SendMessage(vhWnd[i], TBM_SETTICFREQ, 8 , 0);
				SendMessage(vhWnd[i], TBM_SETPAGESIZE, 0, 8);
			}
			SendMessage(vhWnd[19], TBM_SETRANGE, (WPARAM)TRUE, MAKELONG(-12, 12));
			SendMessage(vhWnd[19], TBM_SETTICFREQ, 4 , 0);
			SendMessage(vhWnd[i], TBM_SETPAGESIZE, 0, 4);
			for (i = 0; i < 6; i++) 
				SendMessage(GetDlgItem(hdlg, IDC_HUE + i), BM_SETCHECK, ((fufl[mfd->nmfunc]& (0x01 << i)) > 0), 0);
			SendMessage(GetDlgItem(hdlg, IDC_ONOFF), BM_SETCHECK, mfd->todoFlg & 0x01, 0);
			SendMessage(GetDlgItem(hdlg, IDC_LUMA), CB_SETCURSEL, (WPARAM)mfd->Y, 0);
			SendMessage(GetDlgItem(hdlg, IDC_PRESET), CB_SETCURSEL, (WPARAM)-1, 0);
			ShowWindow(GetDlgItem(hdlg, IDC_KEEPL), winst[keponf[mfd->nmfunc]]);
			SendMessage(GetDlgItem(hdlg, IDC_VERBUTTON), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)txt1[mfd->nmfunc]);
			SendMessage(GetDlgItem(hdlg, IDC_HORSLADER), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)txt2[mfd->nmfunc]);
			for (i=0;i<20;i++) {
				long s = ((i>0) && (i<18)) ? -1 : 1;
				SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, s * mfd->lv[mfd->nmfunc][i]);
			}
			for (i=0;i<17;i++) {
				int enbl = (mfd->Flg[mfd->nmfunc] >> i) & 0x01;
				ShowWindow(vhWnd[i+1], winst[enbl]);
				SendMessage(GetDlgItem(hdlg, IDC_FLV00 + i), BM_SETCHECK, enbl, 0);
			}
			SendMessage(GetDlgItem(hdlg, IDC_KEEPL), BM_SETCHECK, ((mfd->Flg[mfd->nmfunc] & 0x020000) > 0), 0);
			SendMessage(GetDlgItem(hdlg, IDC_LOCK), BM_SETCHECK, ((mfd->Flg[mfd->nmfunc] & 0x040000) > 0), 0);
			SetDlgItemText(hdlg, IDC_MESS, nult);
			SetDlgItemInt(hdlg, IDC_PEAKPOS, mfd->lv[mfd->nmfunc][19], TRUE);
			SetDlgItemText(hdlg, IDC_ONOFF, txt3[mfd->todoFlg & 0x01]);

			mfd->ifp->InitButton(GetDlgItem(hdlg, IDPREVIEW));
			mfd->ifp->Toggle(hdlg);
			return TRUE;

		case WM_COMMAND:
			switch(LOWORD(wParam)) {
			case IDPREVIEW:
				mfd->ifp->Toggle(hdlg);
				break;
			case IDOK:
				EndDialog(hdlg, 0);
				return TRUE;
			case IDHELP:
				{
				char prog[256];
				char path[256];
				LPTSTR ptr;
				GetModuleFileName(NULL, prog, 255);
				GetFullPathName(prog, 255, path, &ptr);
				*ptr = 0;
				strcat(path, "plugins\\rgbeq.hlp");
				ShellExecute(hdlg, "open", path, NULL, NULL, SW_SHOWNORMAL);
				return TRUE;
				}
			case IDCANCEL:
				EndDialog(hdlg, 1);
				return TRUE;
			case IDC_ZGAMMA:
				if(!((mfd->Flg[mfd->nmfunc] & 0x040000) > 0)){
					mfd->lv[mfd->nmfunc][0] = 0;
					SendMessage(vhWnd[0], TBM_SETPOS, (WPARAM)TRUE, mfd->lv[mfd->nmfunc][0]);
					SetDlgItemInt(hdlg, IDC_MESS, 0, TRUE);
				}
				setFlg(mfd, 0);
				mfd->ifp->RedoFrame();
				return TRUE;
			case IDC_ZEQ:
				if(!((mfd->Flg[mfd->nmfunc] & 0x040000) > 0)){
					for (i=1;i<18;i++) {
						mfd->lv[mfd->nmfunc][i] = 0;
						SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, mfd->lv[mfd->nmfunc][i]);
						SetDlgItemInt(hdlg, IDC_MESS, 0, TRUE);
					}
				}
				setFlg(mfd, 20);
				mfd->ifp->RedoFrame();
				return TRUE;
			case IDC_ZPF:
				if(!((mfd->Flg[mfd->nmfunc] & 0x040000) > 0)){
					mfd->lv[mfd->nmfunc][18] = 0;
					SendMessage(vhWnd[18], TBM_SETPOS, (WPARAM)TRUE, mfd->lv[mfd->nmfunc][18]);
					SetDlgItemInt(hdlg, IDC_MESS, 0, TRUE);
				}
				setFlg(mfd, 18);
				mfd->ifp->RedoFrame();
				return TRUE;
			case IDC_ZPEAKP:
				if(!((mfd->Flg[mfd->nmfunc] & 0x040000) > 0)){
					mfd->lv[mfd->nmfunc][19] = 0;
					SendMessage(vhWnd[19], TBM_SETPOS, (WPARAM)TRUE, mfd->lv[mfd->nmfunc][19]);
					SetDlgItemInt(hdlg, IDC_PEAKPOS, 0, TRUE);
					SetDlgItemInt(hdlg, IDC_MESS, 0, TRUE);
				}
				setFlg(mfd, 19);
				mfd->ifp->RedoFrame();
				return TRUE;
			case IDC_ZALL:
				{
				int k,l;
				for (k=0; k<10; k++)
					for (l=0; l<20; l++)
						mfd->lv[k][l] = 0;
				for (k=0;k<10;k++)
					mfd->Flg[k] &= 0xfff9ffff;
				mfd->todoFlg = 0x01;
				}				
				for (i=0;i<20;i++) 
					SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, mfd->lv[mfd->nmfunc][i]);
				SetDlgItemInt(hdlg, IDC_PEAKPOS, 0, TRUE);
				SetDlgItemInt(hdlg, IDC_MESS, 0, TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_KEEPL), BM_SETCHECK, 0, 0);
				SendMessage(GetDlgItem(hdlg, IDC_LOCK), BM_SETCHECK, 0, 0);
				setFlg(mfd, 255);
				mfd->ifp->RedoFrame();
				return TRUE;
			case IDC_RC:
				if(mfd->nmfunc == 1) SendMessage(GetDlgItem(hdlg, IDC_RC), BM_SETCHECK, TRUE, 0);
			case IDC_GC:
				if(mfd->nmfunc == 2) SendMessage(GetDlgItem(hdlg, IDC_GC), BM_SETCHECK, TRUE, 0);
			case IDC_BC:
				if(mfd->nmfunc == 3) SendMessage(GetDlgItem(hdlg, IDC_BC), BM_SETCHECK, TRUE, 0);

				SendMessage(GetDlgItem(hdlg, IDC_HUE), BM_SETCHECK, FALSE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_SAT), BM_SETCHECK, FALSE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_SHARP), BM_SETCHECK, FALSE, 0);

				nf = IsDlgButtonChecked(hdlg, IDC_RC) << 1 |
					 IsDlgButtonChecked(hdlg, IDC_GC) << 2 |
					 IsDlgButtonChecked(hdlg, IDC_BC) << 3 ;
				for (i= 1; i<8; i++ ) 
					if(nf == fufl[i]) mfd->nmfunc = i;
				SendMessage(GetDlgItem(hdlg, IDC_PRESET), CB_SETCURSEL, (WPARAM)-1, 0);
				ShowWindow(GetDlgItem(hdlg, IDC_KEEPL), winst[keponf[mfd->nmfunc]]);
				SendMessage(GetDlgItem(hdlg, IDC_VERBUTTON), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)txt1[mfd->nmfunc]);
				SendMessage(GetDlgItem(hdlg, IDC_HORSLADER), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)txt2[mfd->nmfunc]);
				for (i=0;i<20;i++) {
					long s = ((i>0) && (i<18)) ? -1 : 1;
					SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, s * mfd->lv[mfd->nmfunc][i]);
				}
				for (i=0;i<17;i++) {
					int enbl = (mfd->Flg[mfd->nmfunc] >> i) & 0x01;
					ShowWindow(vhWnd[i+1], winst[enbl]);
					SendMessage(GetDlgItem(hdlg, IDC_FLV00 + i), BM_SETCHECK, enbl, 0);
				}
				SendMessage(GetDlgItem(hdlg, IDC_KEEPL), BM_SETCHECK, ((mfd->Flg[mfd->nmfunc] & 0x020000) > 0), 0);
				SendMessage(GetDlgItem(hdlg, IDC_LOCK), BM_SETCHECK, ((mfd->Flg[mfd->nmfunc] & 0x040000) > 0), 0);
				SetDlgItemText(hdlg, IDC_MESS, nult);
				SetDlgItemInt(hdlg, IDC_PEAKPOS, mfd->lv[mfd->nmfunc][19], TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_LUMA), CB_SETCURSEL, (WPARAM)mfd->Y, 0);

				return TRUE;
			case IDC_HUE:
			case IDC_SAT:
			case IDC_SHARP:
				SendMessage(GetDlgItem(hdlg, IDC_RC), BM_SETCHECK, FALSE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_GC), BM_SETCHECK, FALSE, 0);
				SendMessage(GetDlgItem(hdlg, IDC_BC), BM_SETCHECK, FALSE, 0);
				if(IsDlgButtonChecked(hdlg, IDC_HUE)) mfd->nmfunc = 0;
				if(IsDlgButtonChecked(hdlg, IDC_SAT)) mfd->nmfunc = 8;
				if(IsDlgButtonChecked(hdlg, IDC_SHARP)) mfd->nmfunc = 9;
				SendMessage(GetDlgItem(hdlg, IDC_PRESET), CB_SETCURSEL, (WPARAM)-1, 0);
				ShowWindow(GetDlgItem(hdlg, IDC_KEEPL), winst[keponf[mfd->nmfunc]]);
				SendMessage(GetDlgItem(hdlg, IDC_VERBUTTON), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)txt1[mfd->nmfunc]);
				SendMessage(GetDlgItem(hdlg, IDC_HORSLADER), WM_SETTEXT, 0, (LPARAM)(LPCTSTR)txt2[mfd->nmfunc]);
				for (i=0;i<20;i++) {
					long s = ((i>0) && (i<18)) ? -1 : 1;
					SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, s * mfd->lv[mfd->nmfunc][i]);
				}
				for (i=0;i<17;i++) {
					long enbl = (mfd->Flg[mfd->nmfunc] >> i) & 0x01;
					ShowWindow(vhWnd[i+1], winst[enbl]);
					SendMessage(GetDlgItem(hdlg, IDC_FLV00 + i), BM_SETCHECK, enbl, 0);
				}
				SendMessage(GetDlgItem(hdlg, IDC_KEEPL), BM_SETCHECK, ((mfd->Flg[mfd->nmfunc] & 0x020000) > 0), 0);
				SendMessage(GetDlgItem(hdlg, IDC_LOCK), BM_SETCHECK, ((mfd->Flg[mfd->nmfunc] & 0x040000) > 0), 0);
				SetDlgItemText(hdlg, IDC_MESS, nult);
				SetDlgItemInt(hdlg, IDC_PEAKPOS, mfd->lv[mfd->nmfunc][19], TRUE);
				SendMessage(GetDlgItem(hdlg, IDC_LUMA), CB_SETCURSEL, (WPARAM)mfd->Y, 0);

				return TRUE;
			case IDC_KEEPL:
				if(IsDlgButtonChecked(hdlg, IDC_KEEPL)) mfd->Flg[mfd->nmfunc] |= 0x020000;
				else mfd->Flg[mfd->nmfunc] &= ~0x020000;
				setFlg(mfd, 20);
				mfd->ifp->RedoFrame();
				return TRUE;
			case IDC_LOCK:
				if (IsDlgButtonChecked(hdlg, IDC_LOCK)) mfd->Flg[mfd->nmfunc] |= 0x040000;
				else  mfd->Flg[mfd->nmfunc] &= ~0x040000;
				return TRUE;
			case IDC_FLV00:
			case IDC_FLV01:
			case IDC_FLV02:
			case IDC_FLV03:
			case IDC_FLV04:
			case IDC_FLV05:
			case IDC_FLV06:
			case IDC_FLV07:
			case IDC_FLV08:
			case IDC_FLV09:
			case IDC_FLV0A:
			case IDC_FLV0B:
			case IDC_FLV0C:
			case IDC_FLV0D:
			case IDC_FLV0E:
			case IDC_FLV0F:
			case IDC_FLV10:
				SendMessage(GetDlgItem(hdlg, IDC_PRESET), CB_SETCURSEL, (WPARAM)-1, 0);
				i = LOWORD(wParam) - IDC_FLV00;
				if(IsDlgButtonChecked(hdlg, IDC_LOCK)) {
					int enbl = (mfd->Flg[mfd->nmfunc] >> i) & 0x01;
					SendMessage((HWND)lParam, BM_SETCHECK, enbl, 0);
					return TRUE;
				}
				if (IsDlgButtonChecked(hdlg, LOWORD(wParam))) mfd->Flg[mfd->nmfunc] |= (0x00000001 << i);
				else  mfd->Flg[mfd->nmfunc] &= ~(0x00000001 << i);
				mfd->lv[mfd->nmfunc][i+1] = 0;
				SendMessage(vhWnd[i+1], TBM_SETPOS, (WPARAM)TRUE, mfd->lv[mfd->nmfunc][i+1]);
				ShowWindow(vhWnd[i+1], winst[(mfd->Flg[mfd->nmfunc] >> i) & 0x01]);
				setFlg(mfd, i+1);
				mfd->ifp->RedoFrame();
				return TRUE;
			case	IDC_ONOFF:
				if (IsDlgButtonChecked(hdlg, LOWORD(wParam))) mfd->todoFlg |= 0x01;
				else  mfd->todoFlg &= ~0x00000001;
				SetDlgItemText(hdlg, IDC_ONOFF, txt3[mfd->todoFlg & 0x01]);
				mfd->ifp->RedoFrame();
				return TRUE;
			case	IDC_LUMA:
				if(HIWORD(wParam)==CBN_SELCHANGE) {
					mfd->Y = SendMessage((HWND)lParam, CB_GETCURSEL, 0, 0);
					setFlg(mfd, 255);
					mfd->ifp->RedoFrame();
				}
				return TRUE;
			case	IDC_PRESET:

				if(HIWORD(wParam)==CBN_SELCHANGE) {
					if(IsDlgButtonChecked(hdlg, IDC_LOCK)) return TRUE;
					long mask = eqsf[SendMessage((HWND)lParam, CB_GETCURSEL, 0, 0)];
					for (i=0;i<17;i++) {
						long enbl = (mask >> i) & 0x01;
						mfd->lv[mfd->nmfunc][i+1] = 0;
						SendMessage(vhWnd[i+1], TBM_SETPOS, (WPARAM)TRUE, mfd->lv[mfd->nmfunc][i+1]);
						ShowWindow(vhWnd[i+1], winst[enbl]);
						SendMessage(GetDlgItem(hdlg, IDC_FLV00 + i), BM_SETCHECK, enbl, 0);
					}
					mfd->Flg[mfd->nmfunc] = (mfd->Flg[mfd->nmfunc] | mask) & (0x7FFE0000 | mask); 
					setFlg(mfd, 255);
					mfd->ifp->RedoFrame();
				}
				return TRUE;
			}
			break;
		case WM_HSCROLL:
		case WM_VSCROLL: {
			for (i=0;i<20;i++) if((HWND)lParam == vhWnd[i]) break;
			long s = ((i>0) && (i<18)) ? -1 : 1;
			if(!((mfd->Flg[mfd->nmfunc] & 0x040000) > 0)){
				mfd->lv[mfd->nmfunc][i] = s * SendMessage((HWND)lParam, TBM_GETPOS, 0, 0);
				SetDlgItemInt(hdlg, IDC_MESS, mfd->lv[mfd->nmfunc][i], TRUE);
				if(i==19) SetDlgItemInt(hdlg, IDC_PEAKPOS, mfd->lv[mfd->nmfunc][i], TRUE);
				setFlg(mfd, i);
				mfd->ifp->RedoFrame();
			}
			else SendMessage(vhWnd[i], TBM_SETPOS, (WPARAM)TRUE, s * mfd->lv[mfd->nmfunc][i]);
						 }
			break;
	}

	return FALSE;
}

int ConfigProc(FilterActivation *fa, const FilterFunctions *ff, HWND hwnd)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
	MyFilterData mfd_old = *mfd;
	int ret;

	mfd->ifp = fa->ifp;

	if (DialogBoxParam(fa->filter->module->hInstModule, MAKEINTRESOURCE(IDD_FILTER),
		hwnd, ConfigDlgProc, (LPARAM)mfd))
	{
		*mfd = mfd_old;
		ret = TRUE;
	}
    else
	{
		ret = FALSE;
	}
	return(ret);
}

void StringProc(const FilterActivation *fa, const FilterFunctions *ff, char *str)
{
	MyFilterData *mfd = (MyFilterData *)fa->filter_data;
}
 

inline UCHAR mU(long r, long g, long b, long n) {
	switch (n) {
	case 0:
		return (UCHAR)((5 * r + 8 * g + 3 * b)>>4);
	case 1:
		return (UCHAR)((21845 * r + 21846 * g + 21845 * b)>>16);
	case 2:
		return (UCHAR)((19595 * r + 38470 * g +  7471 * b)>>16);
	case 3:
		return (UCHAR)((13933 * r + 46871 * g +  4732 * b)>>16);
	case 4:
		return (UCHAR)(max3(r,g,b));
	case 5:
		return (UCHAR)((max3(r,g,b)+min3(r,g,b))>>1);
	case 6:
		return (UCHAR)((5 * r + 8 * g + 3 * b)>>4);
	default:
		return (UCHAR)(0);
	}
}

inline void keepU(long *r, long *g, long *b, UCHAR u, long n) {
	long dU = (long)u - (long)mU(*r, *g, *b, n);
	*r += dU;
	*g += dU;
	*b += dU;
	normrgb(r, g, b);
}

void setFlg (MyFilterData *mfd, long ns) {
	int i, j, k, l, sct, fn;
	int snum[17];
	int nf = mfd->nmfunc;
	int dP = mfd->lv[nf][19];
	static long pform[10][512];

	if(ns==18 || ns==255) { // peaks form
		double pval;
		if(ns==255) {
			k=0;
			l=10;
		}
		else {
			k=nf;
			l=nf+1;
		}
		for (i=k;i<l;i++) {
			pval = (double)(abs(mfd->lv[i][18]))/32.0 + 1.0;
			if(mfd->lv[i][18] < 0) 
				for (j=0;j<512;j++)
					pform[i][511-j] = (long)(512.0 * pow(double(j)/512.0 , pval));
			else
				for (j=0;j<512;j++) 
					pform[i][j] = (long)(512.0 - 512.0 * pow(double(j+1)/512.0 , pval));
		}
	}

	if( ((nf>0)&&(nf<8)&&(ns==0)) || (ns==255) ) { // gamma
		double gr = mfd->lv[7][0] + mfd->lv[1][0] + mfd->lv[5][0] + mfd->lv[6][0];
		double gg = mfd->lv[7][0] + mfd->lv[2][0] + mfd->lv[4][0] + mfd->lv[6][0];
		double gb = mfd->lv[7][0] + mfd->lv[3][0] + mfd->lv[4][0] + mfd->lv[5][0];
		gr = gr > 128.0 ? 128.0 : gr;
		gg = gg > 128.0 ? 128.0 : gg;
		gb = gb > 128.0 ? 128.0 : gb;
		gr = gr < -128.0 ? -128.0 : gr;
		gg = gg < -128.0 ? -128.0 : gg;
		gb = gb < -128.0 ? -128.0 : gb;
		gr /= 128.0; 
		gg /= 128.0; 
		gb /= 128.0; 

		if (gr == 0.0)
			for (i = 0; i < 256; i++) 
				mfd->rtbl[i] = i;
		else
			for (i = 0; i < 256; i++) {
				if (gr < 0.0) mfd->rtbl[i] = (UCHAR)(255.0 * pow(double(i)/255.0 , 1.0 + 10.0 * pow(-gr, 3.0)));
				else mfd->rtbl[i] = (UCHAR)(255.0 * pow(double(i)/255.0 , 1.0 - 0.96 * pow(gr, 1.8)));
			}
		if (gg == 0.0)
			for (i = 0; i < 256; i++)
				mfd->gtbl[i] = i;
		else 
			for (i = 0; i < 256; i++) {
				if (gg < 0.0) mfd->gtbl[i] = (UCHAR)(255.0 * pow(double(i)/255.0 , 1.0 + 10.0 * pow(-gg, 3.0)));
				else mfd->gtbl[i] = (UCHAR)(255.0 * pow(double(i)/255.0 , 1.0 - 0.96 * pow(gg, 1.8)));
			}
		if (gb == 0.0)
			for (i = 0; i < 256; i++)
				mfd->btbl[i] = i; 
		else 
			for (i = 0; i < 256; i++){
				if (gb < 0.0) mfd->btbl[i] = (UCHAR)(255.0 * pow(double(i)/255.0 , 1.0 + 10.0 * pow(-gb, 3.0)));
				else mfd->btbl[i] = (UCHAR)(255.0 * pow(double(i)/255.0 , 1.0 - 0.96 * pow(gb, 1.8)));
			}
	}

	if(nf==0 || ns==255) { // hue
		fn = 0;
		sct=0;
		int dofl=0;
		for(i=0;i<18;i++) if(mfd->lv[fn][i] != 0) dofl=1;

		if (dofl) {
			mfd->todoFlg |= 0x00000002;
			for(i=0;i<17;i++) {
				snum[i] = 0;
				if((mfd->Flg[fn] >> i) & 0x00000001) {
					snum[sct] = i+1;
					sct++;
				}
			}
			if (sct==0) {
				for(i=0;i<255;i++) mfd->hue[i] = (mfd->lv[fn][0] * 3)>>1;
			}
			else {

				if (sct==1) 
					for(i=0;i<255;i++) {
						mfd->hue[i] = mfd->lv[fn][0] + mfd->lv[fn][snum[0]];
						if(mfd->hue[i] > 128) mfd->hue[i] = 128;
						if(mfd->hue[i] < -128) mfd->hue[i] = -128;
						mfd->hue[i] = (mfd->hue[i] * 3)>>1;
					}
				else { 
					if(snum[0]>1) {
						int pose = ((snum[0] - 1)<<4) + mfd->lv[fn][19];
						for(i=0;i < pose ;i++) {
							mfd->hue[i] = mfd->lv[fn][0] + ((mfd->lv[fn][snum[0]] * pform[fn][511 - (i<<9)/pose])>>9);
							if(mfd->hue[i]>128) mfd->hue[i] = 128;
							if(mfd->hue[i]<-128) mfd->hue[i] = -128;
							mfd->hue[i] = (mfd->hue[i] * 3)>>1;
						}
					}
					for (i=0;i < (sct - 1);i++) {
						int poss = ((snum[i] - 1)<<4) + ((snum[i]>1) ? mfd->lv[fn][19] : 0);
						int rang = ((snum[i+1] - 1)<<4) + ((snum[i+1]<17) ? mfd->lv[fn][19] : 0) - poss;
						for(j=0;j < rang ;j++) {
							mfd->hue[poss + j] = mfd->lv[fn][0] +
							((mfd->lv[fn][snum[i]] * pform[fn][(j<<9)/rang])>>9) +
							((mfd->lv[fn][snum[i+1]] * pform[fn][(511 - (j<<9)/rang)])>>9);
							if(mfd->hue[poss + j]>128) mfd->hue[poss + j] = 128;
							if(mfd->hue[poss + j]<-128) mfd->hue[poss + j] = -128;
							mfd->hue[poss + j] = (mfd->hue[poss + j] * 3)>>1;
						}
					}
					if(snum[sct-1]<16) {
						int poss = ((snum[sct-1]-1)<<4) + mfd->lv[fn][19];
						int rang = 255 - poss;
						for(i=0;i < rang ;i++) {
							mfd->hue[poss + i] = mfd->lv[fn][0] + ((mfd->lv[fn][snum[sct-1]] * pform[fn][(i<<9)/rang])>>9);
							if(mfd->hue[poss + i]>128) mfd->hue[i] = 128;
							if(mfd->hue[poss + i]<-128) mfd->hue[i] = -128;
							mfd->hue[poss + i] = (mfd->hue[poss + i] * 3)>>1;
						}
					}
				}
			}
		}
		else mfd->todoFlg &= ~0x00000002;
	}

	if( ((nf>0)&&(nf<8)&&(ns>0)) || (ns==255) ) { // rgb
		int flg = 0;
		for(i=0;i<256;i++) {
			mfd->incr[i] = 0;
			mfd->incg[i] = 0;
			mfd->incb[i] = 0;
			mfd->incrk[i] = 0;
			mfd->incgk[i] = 0;
			mfd->incbk[i] = 0;
		}

		fn = 7;
		sct=0;
		int dofl=0;
		for(i=1;i<18;i++) if(mfd->lv[fn][i] != 0) dofl=1;
		if (dofl) {
			flg = 1;
			for(i=0;i<17;i++) {
				snum[i] = 0;
				if((mfd->Flg[fn] >> i) & 0x00000001) {
					snum[sct] = i+1;
					sct++;
				}
			}
			if (sct>0) {
				if (sct==1) 
					for(i=0;i<255;i++) {
						mfd->incr[i] += mfd->lv[fn][snum[0]];
						if(mfd->incr[i] > 128) mfd->incr[i] = 128;
						if(mfd->incr[i] < -128) mfd->incr[i] = -128;
						mfd->incg[i] = mfd->incr[i];
						mfd->incb[i] = mfd->incr[i];
					}
				else { 
					if(snum[0]>1) {
						int pose = ((snum[0] - 1)<<4) + mfd->lv[fn][19];
						for(i=0;i < pose ;i++) {
							mfd->incr[i] += 
								((mfd->lv[fn][snum[0]] * pform[fn][511 - (i<<9)/pose])>>9);
							if(mfd->incr[i]>128) mfd->incr[i] = 128;
							if(mfd->incr[i]<-128) mfd->incr[i] = -128;
							mfd->incg[i] = mfd->incr[i];
							mfd->incb[i] = mfd->incr[i];
						}
					}
					for (i=0;i < (sct - 1);i++) {
						int poss = ((snum[i] - 1)<<4) + ((snum[i]>1) ? mfd->lv[fn][19] : 0);
						int rang = ((snum[i+1] - 1)<<4) + ((snum[i+1]<17) ? mfd->lv[fn][19] : 0) - poss;
						for(j=0;j < rang ;j++) {
							mfd->incr[poss + j] += 
								((mfd->lv[fn][snum[i]] * pform[fn][(j<<9)/rang])>>9) +
								((mfd->lv[fn][snum[i+1]] * pform[fn][(511 - (j<<9)/rang)])>>9);
							if(mfd->incr[poss + j]>128) mfd->incr[poss + j] = 128;
							if(mfd->incr[poss + j]<-128) mfd->incr[poss + j] = -128;
							mfd->incg[poss + j] = mfd->incr[poss + j];
							mfd->incb[poss + j] = mfd->incr[poss + j];
						}
					}
					if(snum[sct-1]<17) {
						int poss = ((snum[sct-1]-1)<<4) + mfd->lv[fn][19];
						int rang = 255 - poss;
						for(i=0;i < rang ;i++) {
							mfd->incr[poss + i] += 
								((mfd->lv[fn][snum[sct-1]] * pform[fn][(i<<9)/rang]) >> 9);
							if(mfd->incr[poss + i]>128) mfd->incr[i] = 128;
							if(mfd->incr[poss + i]<-128) mfd->incr[i] = -128;
							mfd->incg[poss + i] = mfd->incr[poss + i];
							mfd->incb[poss + i] = mfd->incr[poss + i];
						}
					}
				}
			}
		}

		for (fn=1;fn<4;fn++) {
			long *tbl1;
			switch(fn) {
			case 1:
				if (mfd->Flg[fn] & 0x00020000) tbl1 = mfd->incrk;
				else  tbl1 = mfd->incr;
				break;
			case 2:
				if (mfd->Flg[fn] & 0x00020000) tbl1 = mfd->incgk;
				else  tbl1 = mfd->incg;
				break;
			case 3:
				if (mfd->Flg[fn] & 0x00020000) tbl1 = mfd->incbk;
				else  tbl1 = mfd->incb;
				break;
			}
			sct=0;
			dofl=0;
			for(i=1;i<18;i++) if(mfd->lv[fn][i] != 0) dofl=1;
			if (dofl) {
				flg = 1;
				for(i=0;i<17;i++) {
					snum[i] = 0;
					if((mfd->Flg[fn] >> i) & 0x00000001) {
						snum[sct] = i+1;
						sct++;
					}
				}
				if (sct>0) {
					if (sct==1) 
						for(i=0;i<255;i++) {
							tbl1[i] += mfd->lv[fn][snum[0]];
							if(tbl1[i] > 128) tbl1[i] = 128;
							if(tbl1[i] < -128) tbl1[i] = -128;
						}
					else { 
						if(snum[0]>1) {
							int pose = ((snum[0] - 1)<<4) + mfd->lv[fn][19];
							for(i=0;i < pose ;i++) {
								tbl1[i] += 
									((mfd->lv[fn][snum[0]] * pform[fn][511 - (i<<9)/pose])>>9);
								if(tbl1[i]>128) tbl1[i] = 128;
								if(tbl1[i]<-128) tbl1[i] = -128;
							}
						}
						for (i=0;i < (sct - 1);i++) {
							int poss = ((snum[i] - 1)<<4) + ((snum[i]>1) ? mfd->lv[fn][19] : 0);
							int rang = ((snum[i+1] - 1)<<4) + ((snum[i+1]<17) ? mfd->lv[fn][19] : 0) - poss;
							for(j=0;j < rang ;j++) {
								tbl1[poss + j] += 
									((mfd->lv[fn][snum[i]] * pform[fn][(j<<9)/rang])>>9) +
									((mfd->lv[fn][snum[i+1]] * pform[fn][(511 - (j<<9)/rang)])>>9);
								if(tbl1[poss + j]>128) tbl1[poss + j] = 128;
								if(tbl1[poss + j]<-128) tbl1[poss + j] = -128;
							}
						}
						if(snum[sct-1]<17) {
							int poss = ((snum[sct-1]-1)<<4) + mfd->lv[fn][19];
							int rang = 255 - poss;
							for(i=0;i < rang ;i++) {
								tbl1[poss + i] += 
									((mfd->lv[fn][snum[sct-1]] * pform[fn][(i<<9)/rang])>>9);
								if(tbl1[poss + i]>128) tbl1[i] = 128;
								if(tbl1[poss + i]<-128) tbl1[i] = -128;
							}
						}
					}
				}
			}
		}

		for (fn=4;fn<7;fn++) {
			long *tbl1;
			long *tbl2;
			long val;
			switch(fn) {
			case 4:
				if (mfd->Flg[fn] & 0x00020000) {
					tbl1 = mfd->incbk;
					tbl2 = mfd->incgk;
				}
				else {
					tbl1 = mfd->incb;
					tbl2 = mfd->incg;
				}
				break;
			case 5:
				if (mfd->Flg[fn] & 0x00020000) {
					tbl1 = mfd->incbk;
					tbl2 = mfd->incrk;
				}
				else {
					tbl1 = mfd->incb;
					tbl2 = mfd->incr;
				}
				break;
			case 6:
				if (mfd->Flg[fn] & 0x00020000) {
					tbl1 = mfd->incrk;
					tbl2 = mfd->incgk;
				}
				else {
					tbl1 = mfd->incr;
					tbl2 = mfd->incg;
				}
				break;
			}
			sct=0;
			dofl=0;
			for(i=1;i<18;i++) if(mfd->lv[fn][i] != 0) dofl=1;
			if (dofl) {
				flg = 1;
				for(i=0;i<17;i++) {
					snum[i] = 0;
					if((mfd->Flg[fn] >> i) & 0x00000001) {
						snum[sct] = i+1;
						sct++;
					}
				}
				if (sct>0) {
					if (sct==1) 
						for(i=0;i<255;i++) {
							val = mfd->lv[fn][snum[0]];
							tbl1[i] += val;
							if(tbl1[i] > 128) tbl1[i] = 128;
							if(tbl1[i] < -128) tbl1[i] = -128;
							tbl2[i] += val;
							if(tbl2[i] > 128) tbl2[i] = 128;
							if(tbl2[i] < -128) tbl2[i] = -128;
						}
					else { 
						if(snum[0]>1) {
							int pose = ((snum[0] - 1)<<4) + mfd->lv[fn][19];
							for(i=0;i < pose ;i++) {
								val = 
									((mfd->lv[fn][snum[0]] * pform[fn][511 - (i<<9)/pose])>>9);
								tbl1[i] += val;
								if(tbl1[i]>128) tbl1[i] = 128;
								if(tbl1[i]<-128) tbl1[i] = -128;
								tbl2[i] += val;
								if(tbl2[i]>128) tbl2[i] = 128;
								if(tbl2[i]<-128) tbl2[i] = -128;
							}
						}
						for (i=0;i < (sct - 1);i++) {
							int poss = ((snum[i] - 1)<<4) + ((snum[i]>1) ? mfd->lv[fn][19] : 0);
							int rang = ((snum[i+1] - 1)<<4) + ((snum[i+1]<17) ? mfd->lv[fn][19] : 0) - poss;
							for(j=0;j < rang ;j++) {
								val = 
									((mfd->lv[fn][snum[i]] * pform[fn][(j<<9)/rang])>>9) +
									((mfd->lv[fn][snum[i+1]] * pform[fn][(511 - (j<<9)/rang)])>>9);
								tbl1[poss + j] += val;
								if(tbl1[poss + j]>128) tbl1[poss + j] = 128;
								if(tbl1[poss + j]<-128) tbl1[poss + j] = -128;
								tbl2[poss + j] += val;
								if(tbl2[poss + j]>128) tbl2[poss + j] = 128;
								if(tbl2[poss + j]<-128) tbl2[poss + j] = -128;
							}
						}
						if(snum[sct-1]<17) {
							int poss = ((snum[sct-1]-1)<<4) + mfd->lv[fn][19];
							int rang = 255 - poss;
							for(i=0;i < rang ;i++) {
								val =
									((mfd->lv[fn][snum[sct-1]] * pform[fn][(i<<9)/rang])>>9);
								tbl1[poss + i] += val;
								if(tbl1[poss + i]>128) tbl1[i] = 128;
								if(tbl1[poss + i]<-128) tbl1[i] = -128;
								tbl2[poss + i] += val;
								if(tbl2[poss + i]>128) tbl2[i] = 128;
								if(tbl2[poss + i]<-128) tbl2[i] = -128;
							}
						}
					}
				}
			}
		}

	if(flg) mfd->todoFlg |= 0x00000004;
	else mfd->todoFlg &= ~0x00000004;
	}

	if(nf==8 || ns==255) { // sat
		fn=8;
		sct=0;
		int dofl=0;
		for(i=0;i<18;i++) if(mfd->lv[fn][i] != 0) dofl=1;

		if (dofl) {
			mfd->todoFlg |= 0x00000008;
			for(i=0;i<17;i++) {
				snum[i] = 0;
				if((mfd->Flg[fn] >> i) & 0x00000001) {
					snum[sct] = i+1;
					sct++;
				}
			}
			if (sct==0) {
				for(i=0;i<255;i++) {
					mfd->incs[i] = mfd->lv[fn][0];
					mfd->incs[i] = 128 + mfd->incs[i] * 2 + abs(mfd->incs[i]);
				}
			}
			else {

				if (sct==1) 
					for(i=0;i<255;i++) {
						mfd->incs[i] = mfd->lv[fn][0] + mfd->lv[fn][snum[0]];
						if(mfd->incs[i] > 128) mfd->incs[i] = 128;
						if(mfd->incs[i] < -128) mfd->incs[i] = -128;
						mfd->incs[i] = 128 + mfd->incs[i] * 2 + abs(mfd->incs[i]);
					}
				else { 
					if(snum[0]>1) {
						int pose = ((snum[0] - 1)<<4) + mfd->lv[fn][19];
						for(i=0;i < pose ;i++) {
							mfd->incs[i] = mfd->lv[fn][0] + ((mfd->lv[fn][snum[0]] * pform[fn][511 - (i<<9)/pose])>>9);
							if(mfd->incs[i]>128) mfd->incs[i] = 128;
							if(mfd->incs[i]<-128) mfd->incs[i] = -128;
							mfd->incs[i] = 128 + mfd->incs[i] * 2 + abs(mfd->incs[i]);
						}
					}
					for (i=0;i < (sct - 1);i++) {
						int poss = ((snum[i] - 1)<<4) + ((snum[i]>1) ? mfd->lv[fn][19] : 0);
						int rang = ((snum[i+1] - 1)<<4) + ((snum[i+1]<17) ? mfd->lv[fn][19] : 0) - poss;
						for(j=0;j < rang ;j++) {
							mfd->incs[poss + j] = mfd->lv[fn][0] +
							((mfd->lv[fn][snum[i]] * pform[fn][(j<<9)/rang])>>9) +
							((mfd->lv[fn][snum[i+1]] * pform[fn][(511 - (j<<9)/rang)])>>9);
							if(mfd->incs[poss + j]>128) mfd->incs[poss + j] = 128;
							if(mfd->incs[poss + j]<-128) mfd->incs[poss + j] = -128;
							mfd->incs[poss + j] = 128 + mfd->incs[poss + j] * 2 + abs(mfd->incs[poss + j]);
						}
					}
					if(snum[sct-1]<16) {
						int poss = ((snum[sct-1]-1)<<4) + mfd->lv[fn][19];
						int rang = 255 - poss;
						for(i=0;i < rang ;i++) {
							mfd->incs[poss + i] = mfd->lv[fn][0] + ((mfd->lv[fn][snum[sct-1]] * pform[fn][(i<<9)/rang])>>9);
							if(mfd->incs[poss + i]>128) mfd->incs[i] = 128;
							if(mfd->incs[poss + i]<-128) mfd->incs[i] = -128;
							mfd->incs[poss + i] = 128 + mfd->incs[poss + i] * 2 + abs(mfd->incs[poss + i]);
						}
					}
				}
			}
		}
		else mfd->todoFlg &= ~0x00000008;
	}

	if(nf==9 || ns==255) { // shsm
		fn = 9;
		sct=0;
		int dofl=0;
		for(i=1;i<18;i++) if(mfd->lv[fn][i] != 0) dofl=1;

		if (dofl) {
			mfd->todoFlg |= 0x00000010;
			for(i=0;i<17;i++) {
				snum[i] = 0;
				if((mfd->Flg[fn] >> i) & 0x00000001) {
					snum[sct] = i+1;
					sct++;
				}
			}
			if (sct>0) {
				if (sct==1) 
					for(i=0;i<255;i++) {
						mfd->shsm[i] = mfd->lv[fn][snum[0]];
						if(mfd->shsm[i] > 128) mfd->shsm[i] = 128;
						if(mfd->shsm[i] < -128) mfd->shsm[i] = -128;
					}
				else { 
					if(snum[0]>1) {
						int pose = ((snum[0] - 1)<<4) + mfd->lv[fn][19];
						for(i=0;i < pose ;i++) {
							mfd->shsm[i] = 
								((mfd->lv[fn][snum[0]] * pform[fn][511 - (i<<9)/pose])>>9);
							if(mfd->shsm[i]>128) mfd->shsm[i] = 128;
							if(mfd->shsm[i]<-128) mfd->shsm[i] = -128;
						}
					}
					for (i=0;i < (sct - 1);i++) {
						int poss = ((snum[i] - 1)<<4) + ((snum[i]>1) ? mfd->lv[fn][19] : 0);
						int rang = ((snum[i+1] - 1)<<4) + ((snum[i+1]<17) ? mfd->lv[fn][19] : 0) - poss;
						for(j=0;j < rang ;j++) {
							mfd->shsm[poss + j] = 
								((mfd->lv[fn][snum[i]] * pform[fn][(j<<9)/rang])>>9) +
								((mfd->lv[fn][snum[i+1]] * pform[fn][(511 - (j<<9)/rang)])>>9);
							if(mfd->shsm[poss + j]>128) mfd->shsm[poss + j] = 128;
							if(mfd->shsm[poss + j]<-128) mfd->shsm[poss + j] = -128;
						}
					}
					if(snum[sct-1]<17) {
						int poss = ((snum[sct-1]-1)<<4) + mfd->lv[fn][19];
						int rang = 255 - poss;
						for(i=0;i < rang ;i++) {
							mfd->shsm[poss + i] = 
								((mfd->lv[fn][snum[sct-1]] * pform[fn][(i<<9)/rang]) >> 9);
							if(mfd->shsm[poss + i]>128) mfd->shsm[i] = 128;
							if(mfd->shsm[poss + i]<-128) mfd->shsm[i] = -128;
						}
					}
				}
			}
		}
		else mfd->todoFlg &= ~0x00000010;
	}
}

inline void normrgb(long *r, long *g, long *b) {
	if (*r < 0) *r = 0;
	if (*r > 255) *r = 255;
	if (*g < 0) *g = 0;
	if (*g > 255) *g = 255;
	if (*b < 0) *b = 0;
	if (*b > 255) *b = 255;
}
